/* scope.c   by Michael Thorpe   2005-07-12 */

#include <stdlib.h>
#include <string.h>
#include "lisp.h"

typedef struct entry entry;

struct entry {
	entry *next;
	obj *name;
	obj *value;
};

#ifdef DEBUG
unsigned long total_scopes=0;
#endif

static scope topscope={1,0,0};
scope *globalscope=0;

static obj undefined_error={&word_objtype,1,{s:{"undefined",9}}};

/* real_lookup doesn't decref/incref anything */
static entry *real_lookup(scope *scope,obj *name) {
	entry *e;

	if(!scope)
		scope=globalscope;
	while(scope) {
		for(e=scope->firstentry;e;e=e->next)
			if(real_eqv(e->name,name))
				return(e);
		scope=scope->parent;
	}
	return(0);
}

obj *define(scope *s,obj *name,obj *value) {
	entry *e;

	e=real_lookup(s,name);
	if(e) {
		decref(e->value);
		e->value=value;
		decscoperef(s);
		return(name);
	}
	incscoperef(globalscope);
	decscoperef(s);
	return(define_local(globalscope,name,value));
}

obj *define_local(scope *s,obj *name,obj *value) {
	entry *e;

	e=(entry *)malloc(sizeof(entry));
	if(!e) {
		decscoperef(s);
		decref(name);
		decref(value);
		return(throwoom());
	}
	e->next=s->firstentry;
	e->name=incref(name);
	e->value=value;
	s->firstentry=e;
	decscoperef(s);
	return(name);
}

void drop_scope(scope *s) {
	scope *t;
	entry *e,*f;

#ifdef DEBUG
	total_scopes--;
#endif
	e=s->firstentry;
	while(e) {
		f=e;
		e=e->next;
		decref(f->name);
		decref(f->value);
		free(f);
	}
	t=s->parent;
#ifdef DEBUG
	memset(s,'s',sizeof(scope));
#else
	free(s);
#endif
	decscoperef(t);
}

obj *lookup(scope *scope,obj *name) {
	entry *e;
	obj *val;

	e=real_lookup(scope,name);
	if(!e) {
		decscoperef(scope);
		return(throw(incref(&undefined_error),name));
	}
	decref(name);
	val=incref(e->value);
	decscoperef(scope);
	return(val);
}

scope *newscope(scope *parent) {
	scope *new;

	if(!parent)
		parent=incscoperef(globalscope);
	new=(scope *)malloc(sizeof(scope));
	if(!new) {
		decscoperef(parent);
		return(0);
	}
#ifdef DEBUG
	total_scopes++;
#endif
	new->refcount=1;
	new->parent=parent;
	new->firstentry=0;
	return(new);
}

int store_builtin(unsigned int numargs,unsigned int wantscope,const char *name,obj *(*rawfunc)()) {
	obj *key,*func;

	key=newword(name,strlen(name));
	if(!key)
		return(-1);
	func=newobj(&rawfunc_objtype);
	if(!func) {
		decref(key);
		return(-1);
	}
	func->value.rawfunc.numargs=numargs;
	func->value.rawfunc.wantscope=wantscope;
	func->value.rawfunc.rawfuncptr.ns0=rawfunc;
	key=define_local(incscoperef(globalscope),key,func);
	if(!key)
		return(-1);
	decref(key);
	return(0);
}

int store_object(const char *name,obj *o) {
	obj *key;

	key=newword(name,strlen(name));
	if(!key)
		return(-1);
	key=define_local(incscoperef(globalscope),key,o);
	if(!key)
		return(-1);
	decref(key);
	return(0);
}

int scope_init() {
	if(!(globalscope=newscope(incscoperef(&topscope)))
	|| store_builtin_s2("#define",define)
	|| store_builtin_s2("#define-local",define_local))
		return(1);
	return(0);
}

