/* stream.c   by Michael Thorpe   2005-07-13 */

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h> /* Just for isatty(3) */
#include "lisp.h"

obj *eoferror,*ioerror,*stderrobj;

static obj *close_file(obj *stream) {
	if(stream->type != &stream_objtype)
		return(throwtypeerror(stream));
	if(!stream->value.stream.closable)
		return(throwaccesserror(stream));
	if(fclose(stream->value.stream.f))
		return(throw(incref(ioerror),stream));
	stream->value.stream.readable=0;
	stream->value.stream.writable=0;
	stream->value.stream.closable=0;
	return(stream);
}

static obj *isatty_func(obj *stream) {
	obj *o;

	if(&stream_objtype != stream->type)
		return(throwtypeerror(stream));
	if(!stream->value.stream.readable && !stream->value.stream.writable)
		return(throwaccesserror(stream));
	if(isatty(fileno(stream->value.stream.f)))
		o=&trueobj;
	else
		o=&nilobj;
	decref(stream);
	return(incref(o));
}

static obj *open_file(obj *name,obj *mode) {
	obj *o;
	int i;
	char n[PATH_MAX+1];
	char m[3];

	if(name->type != &string_objtype) {
		decref(mode);
		return(throwtypeerror(name));
	}
	if(mode->type != &string_objtype || mode->value.s.len>2) {
		decref(name);
		return(throwtypeerror(mode));
	}
	if(name->value.s.len>PATH_MAX) {
		decref(mode);
		return(throwrangecheck(name));
	}
	memcpy(m,mode->value.s.data,mode->value.s.len);
	if('a' != m[0] && 'r' != m[0] && 'w' != m[0]) {
		decref(name);
		return(throwtypeerror(mode));
	}
	if(mode->value.s.len>2) {
		decref(name);
		return(throwargumenterror(mode));
	}
	if(2==mode->value.s.len && '+' != m[1]) {
		decref(name);
		return(throwtypeerror(mode));
	}
	m[mode->value.s.len]='\0';
	for(i=0;i<name->value.s.len;i++) {
		if(!name->value.s.data[i]) {
			decref(mode);
			return(throwargumenterror(name));
		}
		n[i]=name->value.s.data[i];
	}
	n[i]='\0';
	decref(mode);
	o=newobj(&stream_objtype);
	if(!o) {
		decref(name);
		return(0);
	}
	o->value.stream.readable=0;
	o->value.stream.writable=0;
	o->value.stream.closable=0;
	o->value.stream.f=fopen(n,m);
	if(!o->value.stream.f) {
		decref(o);
		return(throwaccesserror(name));
	}
	decref(name);
	if('r'==m[0] || '+'==m[1])
		o->value.stream.readable=1;
	if('r' != m[0] || '+'==m[1])
		o->value.stream.writable=1;
	o->value.stream.closable=1;
	return(o);
}

static obj *sgetc(obj *stream) {
	int c;
	obj *o;

	if(stream->type != &stream_objtype)
		return(throwtypeerror(stream));
	if(!stream->value.stream.readable)
		return(throwaccesserror(stream));
	o=newobj(&integer_objtype);
	if(o) {
		c=real_sgetc(stream);
		if(c==-1) {
			decref(o);
			return(throw(incref(eoferror),stream));
		}
		if(c==-2) {
			decref(o);
			return(throw(incref(ioerror),stream));
		}
		o->value.i=c;
	}
	decref(stream);
	return(o);
}

static obj *speekc(obj *stream) {
	obj *o;
	int c;

	if(stream->type != &stream_objtype)
		return(throwtypeerror(stream));
	if(!stream->value.stream.readable)
		return(throwaccesserror(stream));
	o=newobj(&integer_objtype);
	if(o) {
		c=real_speekc(stream);
		if(c==-1) {
			decref(o);
			return(throw(incref(eoferror),stream));
		}
		if(c==-2) {
			decref(o);
			return(throw(incref(ioerror),stream));
		}
		o->value.i=c;
	}
	decref(stream);
	return(o);
}

static obj *sputc(obj *stream,obj *c) {
	int i;

	if(stream->type != &stream_objtype) {
		decref(c);
		return(throwtypeerror(stream));
	}
	if(c->type != &integer_objtype) {
		decref(stream);
		return(throwtypeerror(c));
	}
	if(!stream->value.stream.writable) {
		decref(c);
		return(throwaccesserror(stream));
	}
	if(c->value.i<0 || c->value.i>UCHAR_MAX) {
		decref(stream);
		return(throwrangecheck(c));
	}
	i=real_sputc(stream,(int)c->value.i);
	if(i==-2) {
		decref(c);
		return(throw(incref(ioerror),stream));
	}
	decref(stream);
	return(c);
}

static void stream_dropobj(obj *in) {
	if(in->value.stream.closable)
		fclose(in->value.stream.f);
}

static obj *stream_tostring(obj *in) {
	obj *out;
	char buf[6+1+2+16+1];
	int i;

	i=snprintf(buf,sizeof(buf),"stream-%p",in->value.stream.f);
	if(i==-1 || i>=sizeof(buf))
		return(throwrangecheck(in));
	decref(in);
	out=newstring(buf,i);
	return(out);
}

static obj *write_string(obj *stream,obj *string) {
	unsigned long i,j;

	if(stream->type != &stream_objtype) {
		decref(string);
		return(throwtypeerror(stream));
	}
	if(string->type != &string_objtype) {
		decref(stream);
		return(throwtypeerror(string));
	}
	if(!stream->value.stream.writable) {
		decref(string);
		return(throwaccesserror(stream));
	}
	for(i=0;i<string->value.s.len;i+=j) {
		j=fwrite(string->value.s.data,1,string->value.s.len,stream->value.stream.f);
		if(-1==j || 0==j) {
			decref(string);
			return(throw(incref(ioerror),stream));
		}
	}
	decref(stream);
	return(string);
}

obj *writeobj(obj *stream,obj *o) {
	obj *s;

	s=(o->type->tostring)(o);
	if(!s) {
		decref(stream);
		return(0);
	}
	return(write_string(stream,s));
}

int real_sgetc(obj *stream) {
	int c;

	c=fgetc(stream->value.stream.f);
	if(c==EOF)
		c=-1;
	return(c);
}

int real_speekc(obj *stream) {
	int c;

	c=fgetc(stream->value.stream.f);
	if(c==EOF)
		return(-1);
	if(EOF==ungetc(c,stream->value.stream.f))
		return(-2);
	return(c);
}

int real_sputc(obj *stream,int c) {
	if(EOF==fputc(c,stream->value.stream.f))
		return(-2);
	return(c);
}

int stream_init() {
	obj *in,*out;

	if(!(eoferror=newword("end-of-file",11))
	|| !(ioerror=newword("io-error",8))
	|| !(in=newobj(&stream_objtype))
	|| !(out=newobj(&stream_objtype))
	|| !(stderrobj=newobj(&stream_objtype))
	|| store_builtin_ns1("close-file",close_file)
	|| store_builtin_ns1("isatty",isatty_func)
	|| store_builtin_ns2("open-file",open_file)
	|| store_builtin_ns1("read-char",sgetc)
	|| store_builtin_ns1("peek-char",speekc)
	|| store_builtin_ns2("write-char",sputc)
	|| store_builtin_ns2("write-string",write_string)
	|| store_builtin_ns2("write-object",writeobj))
		return(1);
	in->value.stream.readable=1;
	in->value.stream.writable=0;
	in->value.stream.closable=0;
	in->value.stream.f=stdin;
	out->value.stream.readable=0;
	out->value.stream.writable=1;
	out->value.stream.closable=0;
	out->value.stream.f=stdout;
	stderrobj->value.stream.readable=0;
	stderrobj->value.stream.writable=1;
	stderrobj->value.stream.closable=0;
	stderrobj->value.stream.f=stderr;
	if(store_object("#standard-in",in)
	|| store_object("#standard-out",out)
	|| store_object("#standard-error",incref(stderrobj)))
		return(1);
	INC_KNOWN_OBJECTS(3);
	return(0);
}

objtype stream_objtype={
	name: "stream",
	tostring: stream_tostring,
	dropobj: stream_dropobj
};

